// Copyright Kamu Data, Inc. and contributors. All rights reserved.
//
// Use of this software is governed by the Business Source License
// included in the LICENSE file.
//
// As of the Change Date specified in that file, in accordance with
// the Business Source License, use of this software will be governed
// by the Apache License, Version 2.0.

use datafusion::parquet::{
    basic::LogicalType,
    file::reader::{FileReader, SerializedFileReader},
    schema::types::Type,
};
use datafusion::{
    catalog::{catalog::CatalogProvider, schema::SchemaProvider},
    datasource::TableProvider,
    prelude::*,
};
use dill::*;
use opendatafabric::*;
use std::{
    path::{Path, PathBuf},
    sync::Arc,
};
use tracing::info_span;

use crate::domain::*;

use super::{utils::datafusion_hacks::ListingTableOfFiles, DatasetLayout, VolumeLayout};

pub struct QueryServiceImpl {
    dataset_reg: Arc<dyn DatasetRegistry>,
    volume_layout: Arc<VolumeLayout>,
}

#[component(pub)]
impl QueryServiceImpl {
    pub fn new(dataset_reg: Arc<dyn DatasetRegistry>, volume_layout: Arc<VolumeLayout>) -> Self {
        Self {
            dataset_reg,
            volume_layout,
        }
    }
}

#[async_trait::async_trait]
impl QueryService for QueryServiceImpl {
    async fn tail(
        &self,
        dataset_ref: &DatasetRefLocal,
        num_records: u64,
    ) -> Result<Arc<dyn DataFrame>, QueryError> {
        let dataset_handle = self.dataset_reg.resolve_dataset_ref(dataset_ref)?;

        let vocab: DatasetVocabulary = self
            .dataset_reg
            .get_metadata_chain(&dataset_handle.as_local_ref())?
            .iter_blocks()
            .find_map(|(_, b)| b.event.into_variant::<SetVocab>())
            .map(|sv| sv.into())
            .unwrap_or_default();

        // TODO: This is a workaround for Arrow not supporting any operations on Decimals yet
        // See: https://github.com/apache/arrow-rs/issues/272
        let mut has_decimal = false;

        // TODO: This is a workaround for Arrow not handling timestamps with explicit timezones.
        // We basically have to re-cast all timestamp fields into timestamps after querying.
        // See:
        // - https://github.com/apache/arrow-datafusion/issues/959
        // - https://github.com/apache/arrow-rs/issues/393
        let schema = self.get_schema(&dataset_handle.as_local_ref()).await?;
        let fields: Vec<String> = match schema {
            Type::GroupType { fields, .. } => fields
                .iter()
                .map(|f| match f.as_ref() {
                    pt @ Type::PrimitiveType { .. } => {
                        if let Some(LogicalType::TIMESTAMP(ts)) = pt.get_basic_info().logical_type()
                        {
                            if ts.is_adjusted_to_u_t_c {
                                return format!(
                                    "CAST(\"{name}\" as TIMESTAMP) as \"{name}\"",
                                    name = pt.get_basic_info().name()
                                );
                            }
                        } else if pt.get_precision() > 0 {
                            has_decimal = true;
                        }
                        format!("\"{}\"", pt.get_basic_info().name())
                    }
                    Type::GroupType { basic_info, .. } => format!("\"{}\"", basic_info.name()),
                })
                .collect(),
            Type::PrimitiveType { .. } => unreachable!(),
        };

        let query = if !has_decimal {
            format!(
                r#"SELECT {fields} FROM "{dataset}" ORDER BY {offset_col} DESC LIMIT {num_records}"#,
                fields = fields.join(", "),
                dataset = dataset_handle.name,
                offset_col = vocab.offset_column.unwrap_or("offset".to_owned()),
                num_records = num_records
            )
        } else {
            format!(
                r#"SELECT {fields} FROM "{dataset}" DESC LIMIT {num_records}"#,
                fields = fields.join(", "),
                dataset = dataset_handle.name,
                num_records = num_records
            )
        };

        self.sql_statement(
            &query,
            QueryOptions {
                datasets: vec![DatasetQueryOptions {
                    dataset_ref: dataset_handle.as_local_ref(),
                    limit: Some(num_records),
                }],
            },
        )
        .await
    }

    async fn sql_statement(
        &self,
        statement: &str,
        options: QueryOptions,
    ) -> Result<Arc<dyn DataFrame>, QueryError> {
        let span = info_span!("Executing SQL query", statement);
        let _span_guard = span.enter();

        let cfg = ExecutionConfig::new()
            .with_information_schema(true)
            .with_default_catalog_and_schema("kamu", "kamu");

        let mut ctx = ExecutionContext::with_config(cfg);

        ctx.register_catalog(
            "kamu",
            Arc::new(KamuCatalog::new(
                self.dataset_reg.clone(),
                self.volume_layout.clone(),
                options,
            )),
        );

        Ok(ctx.sql(statement).await?)
    }

    async fn get_schema(&self, dataset_ref: &DatasetRefLocal) -> Result<Type, QueryError> {
        let dataset_handle = self.dataset_reg.resolve_dataset_ref(dataset_ref)?;
        let metadata_chain = self
            .dataset_reg
            .get_metadata_chain(&dataset_handle.as_local_ref())?;
        let dataset_layout = DatasetLayout::new(&self.volume_layout, &dataset_handle.name);

        let last_data_file = metadata_chain
            .iter_blocks()
            .filter_map(|(h, b)| b.into_data_stream_block().map(|b| (h, b)))
            .filter(|(_, b)| b.event.output_data.is_some())
            .map(|(block_hash, _)| dataset_layout.data_dir.join(block_hash.to_string()))
            .next()
            .expect("Obtaining schema from datasets with no data is not yet supported");

        let file = std::fs::File::open(&last_data_file)?;
        let reader = SerializedFileReader::new(file).map_err(|e| QueryError::internal(e))?;
        Ok(reader.metadata().file_metadata().schema().clone())
    }
}

/////////////////////////////////////////////////////////////////////////////////////////

struct KamuCatalog {
    dataset_reg: Arc<dyn DatasetRegistry>,
    volume_layout: Arc<VolumeLayout>,
    options: QueryOptions,
}

impl KamuCatalog {
    fn new(
        dataset_reg: Arc<dyn DatasetRegistry>,
        volume_layout: Arc<VolumeLayout>,
        options: QueryOptions,
    ) -> Self {
        Self {
            dataset_reg,
            volume_layout,
            options,
        }
    }
}

impl CatalogProvider for KamuCatalog {
    fn as_any(&self) -> &dyn std::any::Any {
        self
    }

    fn schema_names(&self) -> Vec<String> {
        vec!["kamu".to_owned()]
    }

    fn schema(&self, name: &str) -> Option<Arc<dyn datafusion::catalog::schema::SchemaProvider>> {
        if name == "kamu" {
            Some(Arc::new(KamuSchema::new(
                self.dataset_reg.clone(),
                self.volume_layout.clone(),
                self.options.clone(),
            )))
        } else {
            None
        }
    }
}

/////////////////////////////////////////////////////////////////////////////////////////

// TODO: Performance is poor as it essentially reads all data files in the workspace
// and in some cases (like 'show tables') even twice
struct KamuSchema {
    dataset_reg: Arc<dyn DatasetRegistry>,
    volume_layout: Arc<VolumeLayout>,
    options: QueryOptions,
}

impl KamuSchema {
    fn new(
        dataset_reg: Arc<dyn DatasetRegistry>,
        volume_layout: Arc<VolumeLayout>,
        options: QueryOptions,
    ) -> Self {
        Self {
            dataset_reg,
            volume_layout,
            options,
        }
    }

    fn has_data(&self, dataset_handle: &DatasetHandle) -> bool {
        let limit = self.options_for(dataset_handle).and_then(|o| o.limit);
        let files = self.collect_data_files(dataset_handle, limit);

        if files.is_empty() {
            return false;
        }

        !Self::is_nested(files.first().unwrap())
    }

    fn is_nested(file: &Path) -> bool {
        let reader = SerializedFileReader::new(std::fs::File::open(file).unwrap()).unwrap();
        let schema = reader.metadata().file_metadata().schema();
        schema.get_fields().iter().any(|f| f.is_group())
    }

    fn collect_data_files(
        &self,
        dataset_handle: &DatasetHandle,
        limit: Option<u64>,
    ) -> Vec<PathBuf> {
        let dataset_layout = DatasetLayout::new(&self.volume_layout, &dataset_handle.name);

        if let Ok(metadata_chain) = self
            .dataset_reg
            .get_metadata_chain(&dataset_handle.as_local_ref())
        {
            let mut files = Vec::new();
            let mut num_records = 0;

            for (block_hash, slice) in metadata_chain
                .iter_blocks()
                .filter_map(|(h, b)| b.into_data_stream_block().map(|b| (h, b)))
                .filter_map(|(h, b)| b.event.output_data.map(|s| (h, s)))
            {
                num_records += slice.interval.end - slice.interval.start + 1;
                files.push(dataset_layout.data_dir.join(block_hash.to_string()));
                if limit.is_some() && limit.unwrap() <= num_records as u64 {
                    break;
                }
            }

            files
        } else {
            Vec::new()
        }
    }

    fn options_for(&self, dataset_handle: &DatasetHandle) -> Option<&DatasetQueryOptions> {
        for opt in &self.options.datasets {
            let same = match &opt.dataset_ref {
                DatasetRefLocal::ID(id) => *id == dataset_handle.id,
                DatasetRefLocal::Name(name) => *name == dataset_handle.name,
                DatasetRefLocal::Handle(h) => h.id == dataset_handle.id,
            };
            if same {
                return Some(opt);
            }
        }
        None
    }
}

impl SchemaProvider for KamuSchema {
    fn as_any(&self) -> &dyn std::any::Any {
        self
    }

    fn table_names(&self) -> Vec<String> {
        if self.options.datasets.is_empty() {
            self.dataset_reg
                .get_all_datasets()
                .filter(|hdl| self.has_data(hdl))
                .map(|hdl| hdl.name.to_string())
                .collect()
        } else {
            self.options
                .datasets
                .iter()
                .map(|d| d.dataset_ref.to_string())
                .collect()
        }
    }

    fn table_exist(&self, name: &str) -> bool {
        let dataset_handle: Option<DatasetHandle> = try {
            let dataset_name = DatasetName::try_from(name).ok()?;
            self.dataset_reg
                .resolve_dataset_ref(&dataset_name.into())
                .ok()?
        };
        dataset_handle.is_some()
    }

    fn table(&self, name: &str) -> Option<Arc<dyn TableProvider>> {
        let dataset_handle = self
            .dataset_reg
            .resolve_dataset_ref(&DatasetName::try_from(name).unwrap().into())
            .unwrap();
        let limit = self.options_for(&dataset_handle).and_then(|o| o.limit);
        let files = self.collect_data_files(&dataset_handle, limit);

        if files.is_empty() {
            None
        } else {
            // TODO: Datafusion made it difficult to lazily create table providers,
            // so we have to resort to spawning a thread to call an async function
            // See: https://github.com/apache/arrow-datafusion/issues/1792
            let table = std::thread::spawn(move || {
                let runtime = tokio::runtime::Runtime::new().unwrap();
                runtime.block_on(ListingTableOfFiles::new_with_defaults(
                    files
                        .into_iter()
                        .map(|p| p.to_string_lossy().into())
                        .collect(),
                ))
            })
            .join()
            .unwrap()
            .unwrap();

            Some(Arc::new(table))
        }
    }
}
